import logging
import os
from typing import Any, Dict, Mapping, Optional, Tuple, Type, Union

from langchain.pydantic_v1 import BaseModel, Field
from langchain_community.utilities.dalle_image_generator import DallEAPIWrapper
from langchain_community.utils.openai import is_openai_v1
from langchain_core.callbacks import CallbackManagerForToolRun
from langchain_core.pydantic_v1 import BaseModel, Extra, Field, root_validator
from langchain_core.tools import BaseTool
from langchain_core.utils import get_from_dict_or_env, get_pydantic_field_names

from bisheng_langchain.utils.azure_dalle_image_generator import AzureDallEWrapper

logger = logging.getLogger(__name__)


class DallEAPIWrapper(BaseModel):
    """Wrapper for OpenAI's DALL-E Image Generator.

    https://platform.openai.com/docs/guides/images/generations?context=node

    Usage instructions:

    1. `pip install openai`
    2. save your OPENAI_API_KEY in an environment variable
    """

    client: Any  #: :meta private:
    async_client: Any = Field(default=None, exclude=True)  #: :meta private:
    model_name: str = Field(default="dall-e-2", alias="model")
    model_kwargs: Dict[str, Any] = Field(default_factory=dict)
    openai_api_key: Optional[str] = Field(default=None, alias="api_key")
    """Automatically inferred from env var `OPENAI_API_KEY` if not provided."""
    openai_api_base: Optional[str] = Field(default=None, alias="base_url")
    """Base URL path for API requests, leave blank if not using a proxy or service 
        emulator."""
    openai_organization: Optional[str] = Field(default=None, alias="organization")
    """Automatically inferred from env var `OPENAI_ORG_ID` if not provided."""
    # to support explicit proxy for OpenAI
    openai_proxy: Optional[str] = None
    request_timeout: Union[float, Tuple[float, float], Any, None] = Field(default=None, alias="timeout")
    n: int = 1
    """Number of images to generate"""
    size: str = "1024x1024"
    """Size of image to generate"""
    separator: str = "\n"
    """Separator to use when multiple URLs are returned."""
    quality: Optional[str] = "standard"
    """Quality of the image that will be generated"""
    max_retries: int = 2
    """Maximum number of retries to make when generating."""
    default_headers: Union[Mapping[str, str], None] = None
    default_query: Union[Mapping[str, object], None] = None
    # Configure a custom httpx client. See the
    # [httpx documentation](https://www.python-httpx.org/api/#client) for more details.
    http_client: Union[Any, None] = None
    """Optional httpx.Client."""
    http_async_client: Union[Any, None] = None
    """Optional httpx.AsyncClient. Only used for async invocations. Must specify 
        http_client as well if you'd like a custom client for sync invocations."""

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.forbid

    @root_validator(pre=True)
    def build_extra(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """Build extra kwargs from additional params that were passed in."""
        all_required_field_names = get_pydantic_field_names(cls)
        extra = values.get("model_kwargs", {})
        for field_name in list(values):
            if field_name in extra:
                raise ValueError(f"Found {field_name} supplied twice.")
            if field_name not in all_required_field_names:
                logger.warning(
                    f"""WARNING! {field_name} is not default parameter.
                    {field_name} was transferred to model_kwargs.
                    Please confirm that {field_name} is what you intended."""
                )
                extra[field_name] = values.pop(field_name)

        invalid_model_kwargs = all_required_field_names.intersection(extra.keys())
        if invalid_model_kwargs:
            raise ValueError(
                f"Parameters {invalid_model_kwargs} should be specified explicitly. "
                f"Instead they were passed in as part of `model_kwargs` parameter."
            )

        values["model_kwargs"] = extra
        return values

    @root_validator()
    def validate_environment(cls, values: Dict) -> Dict:
        """Validate that api key and python package exists in environment."""
        values["openai_api_key"] = get_from_dict_or_env(values, "openai_api_key", "OPENAI_API_KEY")
        # Check OPENAI_ORGANIZATION for backwards compatibility.
        values["openai_organization"] = (
            values["openai_organization"] or os.getenv("OPENAI_ORG_ID") or os.getenv("OPENAI_ORGANIZATION") or None
        )
        values["openai_api_base"] = values["openai_api_base"] or os.getenv("OPENAI_API_BASE")
        values["openai_proxy"] = get_from_dict_or_env(
            values,
            "openai_proxy",
            "OPENAI_PROXY",
            default="",
        )

        try:
            import openai

        except ImportError:
            raise ImportError("Could not import openai python package. " "Please install it with `pip install openai`.")

        if is_openai_v1():
            client_params = {
                "api_key": values["openai_api_key"],
                "organization": values["openai_organization"],
                "base_url": values["openai_api_base"],
                "timeout": values["request_timeout"],
                "max_retries": values["max_retries"],
                "default_headers": values["default_headers"],
                "default_query": values["default_query"],
                # "http_client": values["http_client"],
            }

            if not values.get("client"):
                sync_specific = {"http_client": values["http_client"]}
                values["client"] = openai.OpenAI(
                    **client_params,
                    **sync_specific,
                ).images
            if not values.get("async_client"):
                async_specific = {"http_client": values["http_async_client"]}
                values["async_client"] = openai.AsyncOpenAI(
                    **client_params,
                    **async_specific,
                ).images
        elif not values.get("client"):
            values["client"] = openai.Image
        else:
            pass
        return values

    def run(self, query: str) -> str:
        """Run query through OpenAI and parse result."""
        if is_openai_v1():
            response = self.client.generate(
                prompt=query,
                n=self.n,
                size=self.size,
                model=self.model_name,
                quality=self.quality,
            )
            image_urls = self.separator.join([item.url for item in response.data])
        else:
            response = self.client.create(prompt=query, n=self.n, size=self.size, model=self.model_name)
            image_urls = self.separator.join([item["url"] for item in response["data"]])

        return image_urls if image_urls else "No image was generated"


class DallEInput(BaseModel):
    query: str = Field(description="Description about image.")


class DallEImageGenerator(BaseTool):

    name: str = "dalle_image_generator"
    description: str = (
        "A wrapper around OpenAI DALL-E API. Useful for when you need to generate images from a text description. Input should be an image description."
    )
    args_schema: Type[BaseModel] = DallEInput
    api_wrapper: DallEAPIWrapper | AzureDallEWrapper

    def _run(
        self,
        query: str,
        run_manager: Optional[CallbackManagerForToolRun] = None,
    ) -> str:
        """Use the tool."""
        return self.api_wrapper.run(query)



"""tianyancha api"""

from __future__ import annotations

import json
import re
from datetime import datetime
from typing import List, Type

from langchain_core.pydantic_v1 import BaseModel, Field
from loguru import logger

from .base import APIToolBase

asciiPattern = re.compile(r'[A-Z0-9* ]')


# 实时数据Class
class Stock(BaseModel):
    name: str
    todayStart: float
    yesterdayEnd: float
    current: float
    changeAmount: float
    changeRate: float
    vol: float
    turnover: float
    highest: float
    lowest: float
    buyPercent: float

    def __init__(self,
                 name,
                 todayStart,
                 yesterdayEnd,
                 current,
                 highest='0',
                 lowest='0',
                 vol='0',
                 turnover='0'):
        super().__init__(
            name=name,
            todayStart=float(todayStart),
            yesterdayEnd=float(yesterdayEnd),
            current=float(current),
            highest=float(highest),
            lowest=float(lowest),
            changeAmount=round(float(current) - float(yesterdayEnd), 3),
            changeRate=round((float(current) - float(yesterdayEnd)) / float(yesterdayEnd) * 100,
                             3),
            vol=float(vol),
            turnover=float(turnover),
            buyPercent=0.0,
        )

        self.buyPercent = 0.0  # 买卖盘五档委比


class StockArg(BaseModel):
    prefix: str = Field(
        description='前缀。如果是"stock_symbol"传入的为股票代码，则需要传入s_;\
如果"stock_symbol"传入的为指数代码，则为空。',
        default='',
    )
    stock_exchange: str = Field(
        description='交易所简写。股票上市的交易所，或者发布行情指数的交易所。可选项有"sh"(上海证券交易所)、" sz"( 深圳证券交易所)、"bj"( 北京证券交易所)',
    )
    stock_symbol: str = Field(description="""6位数字的股票或者指数代码。
参考信息：
- 如果问题中未给出，可能需要上网查询。
- 上交所股票通常以 6 开头，深交所股票通常以 0、3 开头，北交所股票通常以 8 开头。
- 上交所行情指数通常以 000 开头，深交所指数通常以 399 开头。同一个指数可能会同时在两个交易所发布，例如沪深 300 有"sh000300"和"sz399300"两个代码。""")


stockPattern = re.compile(
    r'var hq_str_s[hz]\d{6}="([^,"]+),([^,"]+),([^,"]+),([^,"]+),([^,"]+),([^,"]+),([^,"]+),([^,"]+),([^,"]+),([^,"]+),[^"]+";'
)
kLinePattern = re.compile(r'var _s[hz]\d{6}_\d+_\d+=\((\[.*?\])\)')


# 计算买卖委比
class StockInfo(APIToolBase):
    'sina stock information'
    args_schema: Type[BaseModel] = StockArg

    def validate_stockNumber(self, stocks: List[str]):
        stockList = []
        for stockNumber in stocks:
            if len(stockNumber) == 8:
                # 8位长度的代码必须以sh或者sz开头，后面6位是数字
                if (stockNumber.startswith('sh')
                        or stockNumber.startswith('sz')) and stockNumber[2:8].isdecimal():
                    stockList.append(stockNumber)
            elif len(stockNumber) == 6:
                # 6位长度的代码必须全是数字
                if stockNumber.isdecimal():
                    # 0开头自动补sz，6开头补sh，3开头补sz，否则无效
                    if stockNumber.startswith('0'):
                        stockList.append('sz' + stockNumber)
                    elif stockNumber.startswith('6'):
                        stockList.append('sh' + stockNumber)
                    elif stockNumber.startswith('3'):
                        stockList.append('sz' + stockNumber)
            elif stockNumber == 'sh':
                stockList.append('sh000001')
            elif stockNumber == 'sz':
                stockList.append('sz399001')
            elif stockNumber == 'zx':
                stockList.append('sz399005')
            elif stockNumber == 'cy':
                stockList.append('sz399006')
            elif stockNumber == '300':
                stockList.append('sh000300')

            return list(map(lambda x: f's_{x}', stockList))

    def devideStock(self, content: str) -> List[Stock]:
        match = stockPattern.search(content)
        stock = []
        if match:
            while match:
                stock.append(
                    Stock(match.group(1), match.group(2), match.group(3), match.group(4),
                          match.group(5), match.group(6), match.group(9), match.group(10)))
                match = stockPattern.search(content, match.end())
        else:
            stock = [content]
        return stock

    def run(self, **kwargs):
        prefix = 's_' if kwargs.get('prefix', '') == 's_' else ''
        stock_symbol = kwargs.get('stock_symbol', '')
        stock_exchange = kwargs.get('stock_exchange', '')
        stock_number = ''.join([prefix, stock_exchange, stock_symbol])

        if self.input_key == 'kLine':
            date_format = '%Y-%m-%d'
            date = kwargs.get('date')
            date_obj = datetime.strptime(date, date_format)
            ts = int(datetime.timestamp(date_obj) * 1000)
            stock = f'{stock_number}_240_{ts}'
            count = datetime.today() - date_obj
            url = self.url.format(stockName=stock_number, stock=stock, count=count.days)
            resp = self.client.get(url)
            if resp.status_code != 200:
                logger.info('api_call_fail res={}', resp.text)
            k_data = resp.text
            k_data = kLinePattern.search(k_data)
            if not k_data:
                return '{}'
            data_array = json.loads(k_data.group(1))
            for item in data_array:
                if item.get('day') == date:
                    return json.dumps(item)
            return '{}'

        resp = super().run(query=stock_number)
        stock = self.devideStock(resp)[0]
        if isinstance(stock, Stock):
            return json.dumps(stock.__dict__, ensure_ascii=False)
        else:
            return stock

    async def arun(self, **kwargs) -> str:
        prefix = 's_' if kwargs.get('prefix', '') == 's_' else ''
        stock_symbol = kwargs.get('stock_symbol', '')
        stock_exchange = kwargs.get('stock_exchange', '')
        stock_number = ''.join([prefix, stock_exchange, stock_symbol])

        if self.input_key == 'kLine':
            date_format = '%Y-%m-%d'
            date = kwargs.get('date')
            date_obj = datetime.strptime(date, date_format)
            ts = int(datetime.timestamp(date_obj) * 1000)
            stock = f'{stock_number}_240_{ts}'
            count = datetime.today() - date_obj
            url = self.url.format(stockName=stock_number, stock=stock, count=count.days)
            k_data = await self.async_client.aget(url)
            k_data = kLinePattern.search(k_data)
            if not k_data:
                return '{}'
            data_array = json.loads(k_data.group(1))
            for item in data_array:
                if item.get('day') == date:
                    return json.dumps(item)

            return '{}'
        else:
            resp = await super().arun(query=stock_number)
            stock = self.devideStock(resp)[0]
            if isinstance(stock, Stock):
                return json.dumps(stock.__dict__, ensure_ascii=False)
            else:
                return stock

    @classmethod
    def realtime_info(cls) -> StockInfo:
        """查询中国A股（沪深北交易所）股票或指数的实时行情数据，返回收盘价、涨跌额、涨跌幅、成交量、成交额"""
        url = 'https://hq.sinajs.cn'
        input_key = 'list'
        headers = {'Referer': 'http://finance.sina.com.cn'}
        return cls(url=url, input_key=input_key, headers=headers)

    @classmethod
    def history_KLine(cls) -> StockInfo:
        """查询中国A股（沪深北交易所）股票或指数的的历史行情数据，返回时间、开盘价、最高价、最低价、收盘价、成交量（股）"""
        url = 'https://quotes.sina.cn/cn/api/jsonp_v2.php/var%20_{stock}=/CN_MarketDataService.getKLineData?symbol={stockName}&scale=240&ma=no&datalen={count}'  # noqa
        input_key = 'kLine'
        header = {'Referer': 'http://finance.sina.com.cn'}

        class stockK(BaseModel):
            stock_symbol: str = Field(description="""6位数字的股票或者指数代码。
参考信息：
- 如果问题中未给出，可能需要上网查询。
- 上交所股票通常以 6 开头，深交所股票通常以 0、3 开头，北交所股票通常以 8 开头。
- 上交所行情指数通常以 000 开头，深交所指数通常以 399 开头。同一个指数可能会同时在两个交易所发布，例如沪深 300 有"sh000300"和"sz399300"两个代码。""")
            stock_exchange: str = Field(
                description=
                '交易所简写。股票上市的交易所，或者发布行情指数的交易所。可选项有"sh"(上海证券交易所)、" sz"( 深圳证券交易所)、"bj"( 北京证券交易所)', )
            date: str = Field(description='需要查询的时间，按照”2024-03-26“格式，传入日期')

        return cls(url=url, input_key=input_key, headers=header, args_schema=stockK)

    # @classmethod
    # def history_KLine(cls, stockid: str) -> StockInfo:
    #     url = 'http://money.finance.sina.com.cn/quotes_service/api/json_v2.php/CN_MarketData.getKLineData'
    #     input_key = 'list'
    #     header = {'Referer': 'http://finance.sina.com.cn'}
    #     return cls(url=url, input_key=input_key, header=header)



preamble="""You are a helpful assistant.
"""

ASSISTANT_PROMPT_COHERE="""{preamble}|<instruct>|Carefully perform the following instructions, in order, starting each with a new line.
Firstly, You may need to use complex and advanced reasoning to complete your task and answer the question. Think about how you can use the provided tools to answer the question and come up with a high level plan you will execute.
Write 'Plan:' followed by an initial high level plan of how you will solve the problem including the tools and steps required.
Secondly, Carry out your plan by repeatedly using actions, reasoning over the results, and re-evaluating your plan. Perform Action, Observation, Reflection steps with the following format. Write 'Action:' followed by a json formatted action containing the "tool_name" and "parameters"
 Next you will analyze the 'Observation:', this is the result of the action.
After that you should always think about what to do next. Write 'Reflection:' followed by what you've figured out so far, any changes you need to make to your plan, and what you will do next including if you know the answer to the question.
... (this Action/Observation/Reflection can repeat N times)
Thirdly, Decide which of the retrieved documents are relevant to the user's last input by writing 'Relevant Documents:' followed by comma-separated list of document numbers. If none are relevant, you should instead write 'None'.
Fourthly, Decide which of the retrieved documents contain facts that should be cited in a good answer to the user's last input by writing 'Cited Documents:' followed a comma-separated list of document numbers. If you dont want to cite any of them, you should instead write 'None'.
Fifthly, Write 'Answer:' followed by a response to the user's last input. Use the retrieved documents to help you. Do not insert any citations or grounding markup.
Finally, Write 'Grounded answer:' followed by a response to the user's last input in high quality natural english. Use the symbols <co: doc> and </co: doc> to indicate when a fact comes from a document in the search result, e.g <co: 4>my fact</co: 4> for a fact from document 4.

Additional instructions to note:
- If the user's question is in Chinese, please answer it in Chinese. 
- 当问题中有涉及到时间信息时，比如最近6个月、昨天、去年等，你需要用时间工具查询时间信息。
""".format(preamble=preamble)


import httpx
from langchain.chat_models import ChatOpenAI
from bisheng_langchain.chat_models import ChatQWen
from langchain.chains import LLMChain
from langchain.prompts.chat import (
    ChatPromptTemplate,
    SystemMessagePromptTemplate,
    HumanMessagePromptTemplate,
)

system_template = """你是一个可靠标题生成或者提取助手。你会收到一篇文档的主要内容，请根据这些内容生成或者提取这篇文档的标题。"""
human_template = """
文档内容如下：
{context}

生成或提取的标题：
"""

messages = [
        SystemMessagePromptTemplate.from_template(system_template),
        HumanMessagePromptTemplate.from_template(human_template),
    ]
title_extract_prompt = ChatPromptTemplate.from_messages(messages)


def extract_title(llm, text, max_length=7000) -> str:
    chain = LLMChain(llm=llm, prompt=title_extract_prompt)
    ans = chain.run(context=text[:max_length])
    return ans 


if __name__ == '__main__':
    llm = ChatQWen(model_name='qwen1.5-72b-chat', 
                   api_key='', 
                   temperature=0.01)
    text = "江苏蔚蓝锂芯股份有限公司\n2021 年年度报告 \n2022 年 03 月\n\n 第一节 重要提示、目录和释义\n公司董事会、监事会及董事、监事、高级管理人员保证年度报告内容的真实、准确、完整，不存在虚假记载、误导性陈述或重大遗漏，并承担个别和连带的法律责任。\n公司负责人 CHEN KAI、主管会计工作负责人林文华及会计机构负责人(会计主管人员)张宗红声明：保证本年度报告中财务报告的真实、准确、完整。\n所有董事均已出席了审议本报告的董事会会议。"
    ans = extract_title(llm, text)
    print(ans)


import json
import os
import numpy as np
import torch
from transformers import AutoModelForSequenceClassification, AutoTokenizer

tokenizer = AutoTokenizer.from_pretrained('/home/public/llm/bge-reranker-large')
model = AutoModelForSequenceClassification.from_pretrained('/home/public/llm/bge-reranker-large').to('cuda:2')
model.eval()


def min_edit_distance(a, b):
    dp = [[0 for i in range(len(b) + 1)] for j in range(len(a) + 1)]
    for i in range(len(a) + 1):
        dp[i][0] = i
    for j in range(len(b) + 1):
        dp[0][j] = j
    for i in range(1, len(a) + 1):
        for j in range(1, len(b) + 1):
            if a[i - 1] == b[j - 1]:
                dp[i][j] = dp[i - 1][j - 1]
            else:
                dp[i][j] = min(dp[i - 1][j] + 1, dp[i][j - 1] + 1, dp[i - 1][j - 1] + 1)
    return dp[-1][-1]


def is_matched(text0, text1, thrd=10):
    text0.replace(" ", "").replace("\n", "")
    text1.replace(" ", "").replace("\n", "")
    dist = min_edit_distance(text0, text1)
    if dist < thrd:
        return True
    return False


def match_score(chunk, query):
    """
    rerank模型计算query和chunk的相似度
    """
    pairs = [[query, chunk]]

    with torch.no_grad():
        inputs = tokenizer(pairs, padding=True, truncation=True, return_tensors='pt', max_length=512).to('cuda:2')
        scores = model(**inputs, return_dict=True).logits.view(-1, ).float()
        scores = torch.sigmoid(scores) 
        scores = scores.cpu().numpy()
        
    return scores[0]


def sort_filter_all_chunks_method1(d, th=0.0):
    """
    rerank模型对所有chunk进行排序
    """
    # answer关键词已提取好
    query = d['question']
    all_chunks = d['all_chunks']

    chunk_match_score = []
    for index, chunk in enumerate(all_chunks):
        chunk_text = chunk['text']
        chunk_match_score.append(match_score(chunk_text, query))

    sorted_res = sorted(enumerate(chunk_match_score), key=lambda x: -x[1])
    print('-----------')
    print(sorted_res)
    remain_chunks = [all_chunks[elem[0]] for elem in sorted_res if elem[1] >= th]
    if not remain_chunks:
        remain_chunks = [all_chunks[sorted_res[0][0]]]

    # for index, chunk in enumerate(remain_chunks):
    #     print('query:', query)
    #     print('chunk_text:', chunk['text'])
    #     print('socre:', sorted_res[index][1])
    #     print('***********')

    d['all_chunks'] = remain_chunks


def calc_precision_recall(d):
    """
    计算分数
    """
    d_ves = d["all_chunks"][:10]
    d_es = d["all_chunks"][10:]
    all_chunks = []
    for i in range(len(d_es)):
        all_chunks.append(d_es[i])
        all_chunks.append(d_ves[i])
    all_chunks.extend(d_ves[i+1:])
    d["all_chunks"] = all_chunks
    
    sort_filter_all_chunks_method1(d)
    NCHUNK = len(d["chunks"])
    NCHUNK_ALL = len(d["all_chunks"])
    print('chunks:', NCHUNK, 'all_chunks:', NCHUNK_ALL)

    scores = np.zeros((NCHUNK, NCHUNK_ALL))
    for j in range(NCHUNK):
        for i in range(NCHUNK):
            if d["chunks"][j]["text"] == d["all_chunks"][i]['text']:
                scores[j][i] = 1
            elif abs(len(d["chunks"][j]["text"]) - len(d["all_chunks"][i]['text'])) > 10:
                scores[j][i] = 0
            elif is_matched(d["chunks"][j]["text"], d["all_chunks"][i]['text']):
                scores[j][i] = 1

    N_gt = NCHUNK
    N_all = NCHUNK_ALL
    if NCHUNK != 0:
        N_right = sum(scores.max(axis=1))
    else:
        N_right = 0
    recall = 0 if N_gt == 0 else N_right / N_gt
    return recall, N_gt, N_right, N_all


with open('zhaogushu_retriever_gt_convert_key.json', 'r') as f:
    retriever_gt_list = json.load(f)
mFieldRecall = 0
total_N_gt, total_N_right, total_N_all = 0, 0, 0
D_SCORES = {}
nquestion = 0
for d in retriever_gt_list:
    recall, N_gt, N_right, N_all = calc_precision_recall(d)
    mFieldRecall += recall
    D_SCORES[d["question"]] = {'recall': recall}
    total_N_gt += N_gt
    total_N_right += N_right
    total_N_all += N_all
    nquestion += 1

mFieldRecall = 0 if nquestion == 0 else mFieldRecall / nquestion
mMethodRecall = total_N_right / total_N_gt

print(f'mFieldRecall: {mFieldRecall * 100:.2f} %')
print(f'mMethodRecall: {mMethodRecall * 100:.2f} %')
print(f'total_N_right: {total_N_right}, total_N_gt: {total_N_gt}, total_N_all: {total_N_all}' )
print(f'nquestion: {nquestion}, mean_N_right: {total_N_right / nquestion}, mean_N_gt: {total_N_gt / nquestion}, mean_N_all: {total_N_all / nquestion}')


from .baseline_vector_retriever import BaselineVectorRetriever
from .keyword_retriever import KeywordRetriever
from .mix_retriever import MixRetriever
from .smaller_chunks_retriever import SmallerChunksVectorRetriever



import torch
from transformers import AutoModelForSequenceClassification, AutoTokenizer


class CustomReranker:

    def __init__(self, model_path, device_id='cuda:0', threshold=0.0):
        self.device_id = device_id
        self.threshold = threshold
        self.tokenizer = AutoTokenizer.from_pretrained(model_path)
        self.rank_model = AutoModelForSequenceClassification.from_pretrained(model_path).to(device_id)
        self.rank_model.eval()

    def match_score(self, chunk, query):
        """
        rerank模型计算query和chunk的相似度
        """
        pairs = [[query, chunk]]

        with torch.no_grad():
            inputs = self.tokenizer(pairs, padding=True, truncation=True, return_tensors='pt', max_length=512).to(self.device_id)
            scores = self.rank_model(**inputs, return_dict=True).logits.view(-1, ).float()
            scores = torch.sigmoid(scores) 
            scores = scores.cpu().numpy()
            
        return scores[0]

    def sort_and_filter(self, query, all_chunks):
        """
        rerank模型对所有chunk进行排序
        """
        chunk_match_score = []
        for index, chunk in enumerate(all_chunks):
            chunk_text = chunk.page_content
            chunk_match_score.append(self.match_score(chunk_text, query))

        sorted_res = sorted(enumerate(chunk_match_score), key=lambda x: -x[1])
        remain_chunks = [all_chunks[elem[0]] for elem in sorted_res if elem[1] >= self.threshold]
        if not remain_chunks:
            remain_chunks = [all_chunks[sorted_res[0][0]]]

        # for index, chunk in enumerate(remain_chunks):
        #     print('query:', query)
        #     print('chunk_text:', chunk.page_content)
        #     print('socre:', sorted_res[index][1])
        #     print('***********')

        return remain_chunks


"""Chain for question-answering against a vector database."""
from __future__ import annotations

import inspect
from abc import abstractmethod
from typing import Any, Dict, List, Optional

from langchain_core.callbacks import (
    AsyncCallbackManagerForChainRun,
    CallbackManagerForChainRun,
    Callbacks
)
from langchain_core.prompts import PromptTemplate, BasePromptTemplate, ChatPromptTemplate, HumanMessagePromptTemplate, SystemMessagePromptTemplate
from langchain_core.language_models import BaseLanguageModel
from langchain_core.pydantic_v1 import Extra, Field
from bisheng_langchain.vectorstores import ElasticKeywordsSearch, Milvus

from langchain.chains.base import Chain
from .bisheng_rag_tool import BishengRAGTool


# system_template = """Use the following pieces of context to answer the user's question. 
# If you don't know the answer, just say that you don't know, don't try to make up an answer.
# ----------------
# {context}"""
# messages = [
#     SystemMessagePromptTemplate.from_template(system_template),
#     HumanMessagePromptTemplate.from_template("{question}"),
# ]
# DEFAULT_QA_PROMPT = ChatPromptTemplate.from_messages(messages)


system_template_general = """你是一个准确且可靠的知识库问答助手，能够借助上下文知识回答问题。你需要根据以下的规则来回答问题：
1. 如果上下文中包含了正确答案，你需要根据上下文进行准确的回答。但是在回答前，你需要注意，上下文中的信息可能存在事实性错误，如果文档中存在和事实不一致的错误，请根据事实回答。
2. 如果上下文中不包含答案，就说你不知道，不要试图编造答案。
3. 你需要根据上下文给出详细的回答，不要试图偷懒，不要遗漏括号中的信息，你必须回答的尽可能详细。
"""
human_template_general = """
上下文：
{context}

问题：
{question}
"""
messages_general = [
    SystemMessagePromptTemplate.from_template(system_template_general),
    HumanMessagePromptTemplate.from_template(human_template_general),
]
DEFAULT_QA_PROMPT = ChatPromptTemplate.from_messages(messages_general)


class BishengRetrievalQA(Chain):
    """Base class for question-answering chains."""

    """Chain to use to combine the documents."""
    input_key: str = "query"  #: :meta private:
    output_key: str = "result"  #: :meta private:
    return_source_documents: bool = False
    """Return the source documents or not."""
    bisheng_rag_tool: BishengRAGTool = Field(
        default_factory=BishengRAGTool, description="RAG tool"
    )

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.forbid
        arbitrary_types_allowed = True
        allow_population_by_field_name = True

    @property
    def input_keys(self) -> List[str]:
        """Input keys.

        :meta private:
        """
        return [self.input_key]

    @property
    def output_keys(self) -> List[str]:
        """Output keys.

        :meta private:
        """
        _output_keys = [self.output_key]
        if self.return_source_documents:
            _output_keys = _output_keys + ["source_documents"]
        return _output_keys

    @classmethod
    def from_llm(
        cls,
        llm: BaseLanguageModel,
        vector_store: Milvus,
        keyword_store: ElasticKeywordsSearch,
        QA_PROMPT: ChatPromptTemplate = DEFAULT_QA_PROMPT,
        max_content: int = 15000,
        sort_by_source_and_index: bool = False,
        callbacks: Callbacks = None,
        return_source_documents: bool = False,
        **kwargs: Any,
    ) -> BishengRetrievalQA:
        bisheng_rag_tool = BishengRAGTool(
            vector_store=vector_store, 
            keyword_store=keyword_store,
            llm=llm,
            QA_PROMPT=QA_PROMPT,
            max_content=max_content,
            sort_by_source_and_index=sort_by_source_and_index,
            **kwargs
        )
        return cls(
            bisheng_rag_tool=bisheng_rag_tool,
            callbacks=callbacks,
            return_source_documents=return_source_documents,
            **kwargs,
        )

    def _call(
        self,
        inputs: Dict[str, Any],
        run_manager: Optional[CallbackManagerForChainRun] = None,
    ) -> Dict[str, Any]:
        """Run get_relevant_text and llm on input query.

        If chain has 'return_source_documents' as 'True', returns
        the retrieved documents as well under the key 'source_documents'.

        Example:
        .. code-block:: python

        res = indexqa({'query': 'This is my query'})
        answer, docs = res['result'], res['source_documents']
        """
        question = inputs[self.input_key]
        if self.return_source_documents:
            answer, docs = self.bisheng_rag_tool.run(question, return_only_outputs=False)
            return {self.output_key: answer, "source_documents": docs}
        else:
            answer = self.bisheng_rag_tool.run(question, return_only_outputs=True)
            return {self.output_key: answer}

    async def _acall(
        self,
        inputs: Dict[str, Any],
        run_manager: Optional[AsyncCallbackManagerForChainRun] = None,
    ) -> Dict[str, Any]:
        """Run get_relevant_text and llm on input query.

        If chain has 'return_source_documents' as 'True', returns
        the retrieved documents as well under the key 'source_documents'.

        Example:
        .. code-block:: python

        res = indexqa({'query': 'This is my query'})
        answer, docs = res['result'], res['source_documents']
        """
        question = inputs[self.input_key]

        if self.return_source_documents:
            answer, docs = await self.bisheng_rag_tool.arun(question, return_only_outputs=False)
            return {self.output_key: answer, "source_documents": docs}
        else:
            answer = await self.bisheng_rag_tool.arun(question, return_only_outputs=True)
            return {self.output_key: answer}



"""Chain that runs an arbitrary python function."""
import logging
from typing import Callable, Dict, Optional

import httpx
from autogen import AssistantAgent
from langchain.base_language import BaseLanguageModel

logger = logging.getLogger(__name__)


class AutoGenAssistant(AssistantAgent):
    """Assistant agent, designed to solve a task with LLM.
    """

    DEFAULT_SYSTEM_MESSAGE = """You are a helpful AI assistant.
Solve tasks using your coding and language skills.
In the following cases, suggest python code (in a python coding block) or shell script (in a sh coding block) for the user to execute.
    1. When you need to collect info, use the code to output the info you need, for example, browse or search the web, download/read a file, print the content of a webpage or a file, get the current date/time, check the operating system. After sufficient info is printed and the task is ready to be solved based on your language skill, you can solve the task by yourself.
    2. When you need to perform some task with code, use the code to perform the task and output the result. Finish the task smartly.
Solve the task step by step if you need to. If a plan is not provided, explain your plan first. Be clear which step uses code, and which step uses your language skill.
When using code, you must indicate the script type in the code block. The user cannot provide any other feedback or perform any other action beyond executing the code you suggest. The user can't modify your code. So do not suggest incomplete code which requires users to modify. Don't use a code block if it's not intended to be executed by the user.
If you want the user to save the code in a file before executing it, put # filename: <filename> inside the code block as the first line. Don't include multiple code blocks in one response. Do not ask users to copy and paste the result. Instead, use 'print' function for the output when relevant. Check the execution result returned by the user.
If the result indicates there is an error, fix the error and output the code again. Suggest the full code instead of partial code or code changes. If the error can't be fixed or if the task is not solved even after the code is executed successfully, analyze the problem, revisit your assumption, collect additional info you need, and think of a different approach to try.
When you find an answer, verify the answer carefully. Include verifiable evidence in your response if possible.
Reply "TERMINATE" in the end when everything is done.
    """ # noqa

    def __init__(
        self,
        name: str,
        model_name: Optional[str] = 'gpt-4-0613',  # when llm_flag=True, need to set
        openai_api_key: Optional[str] = '',  # when llm_flag=True, need to set
        openai_api_base: Optional[str] = '',  # when llm_flag=True, need to set
        openai_proxy: Optional[str] = '',  # when llm_flag=True, need to set
        temperature: Optional[float] = 0,  # when llm_flag=True, need to set
        api_type: Optional[str] = None,  # when llm_flag=True, need to set
        api_version: Optional[str] = None,  # when llm_flag=True, need to set
        llm: Optional[BaseLanguageModel] = None,
        system_message: Optional[
            str] = DEFAULT_SYSTEM_MESSAGE,  # agent system message, llm or group chat manage will use # noqa
        is_termination_msg: Optional[Callable[[Dict], bool]] = None,
        **kwargs,
    ):
        is_termination_msg = (is_termination_msg if is_termination_msg is not None else
                              (lambda x: x.get('content') == 'TERMINATE'))
        config_list = [
            {
                'model': model_name,
                'api_key': openai_api_key,
                'api_base': openai_api_base,
                'api_type': api_type,
                'api_version': api_version,
            },
        ]
        if openai_proxy:
            config_list[0]['http_client'] = httpx.Client(proxies=openai_proxy)
            config_list[0]['http_async_client'] = httpx.AsyncClient(proxies=openai_proxy)

        if llm:
            llm_config = llm

        else:
            llm_config = {
                'seed': 42,  # change the seed for different trials
                'temperature': temperature,
                'config_list': config_list,
                'request_timeout': 120,
            }

        super().__init__(
            name,
            llm_config=llm_config,
            system_message=system_message,
            is_termination_msg=is_termination_msg,
            max_consecutive_auto_reply=None,
            human_input_mode='NEVER',
            code_execution_config=False,
        )



from typing import Union, List
from pydantic import BaseModel


class Message(BaseModel):
    role: str
    content: str


class Function(BaseModel):
    name: str
    description: str
    parameters: dict


class ChatInput(BaseModel):
    model: str
    messages: List[Message] = []
    top_p: float = None
    temperature: float = None
    n: int = 1
    stream: bool = False
    stop: str = None
    max_tokens: int = 256
    functions: List[Function] = []
    function_call: str = None


class Choice(BaseModel):
    index: int
    message: Message = None
    finish_reason: str = 'stop'


class Usage(BaseModel):
    prompt_tokens: int = 0
    completion_tokens: int = 0
    total_tokens: int = 0


class ChatOutput(BaseModel):
    status_code: int
    status_message: str = 'success'
    id: str = None
    object: str = None
    model: str = None
    created: int = None
    choices: List[Choice] = []
    usage: Usage = None


class CompletionsInput(BaseModel):
    model: str
    prompt: str
    top_p: float = None
    temperature: float = None
    n: int = 1
    stream: bool = True
    stop: str = None
    max_tokens: int = 256



from typing import Any, Callable, Dict, List, Tuple

# from .eastmoney import
from langchain_core.tools import BaseTool
from mypy_extensions import KwArg

from .flow import FlowTools
from .macro_data import MacroData
from .sina import StockInfo
from .tianyancha import CompanyInfo

# 筛选出类方法
tianyancha_class_methods = [
    method for method in CompanyInfo.__dict__
    if isinstance(CompanyInfo.__dict__[method], classmethod)
]

_TIAN_YAN_CHA_TOOLS: Dict[str, Tuple[Callable[[KwArg(Any)], BaseTool], List[str]]] = {
    f'tianyancha_{name}': (CompanyInfo.get_api_tool, ['api_key'])
    for name in tianyancha_class_methods
}

sina_class_methods = [
    method for method in StockInfo.__dict__ if isinstance(StockInfo.__dict__[method], classmethod)
]

_SINA_TOOLS: Dict[str, Tuple[Callable[[KwArg(Any)], BaseTool], List[str]]] = {
    f'sina_{name}': (StockInfo.get_api_tool, [])
    for name in sina_class_methods
}

macro_class_methods = [
    method for method in MacroData.__dict__ if isinstance(MacroData.__dict__[method], classmethod)
]

_MACRO_TOOLS: Dict[str, Tuple[Callable[[KwArg(Any)], BaseTool], List[str]]] = {
    f'macro_{name}': (MacroData.get_api_tool, [])
    for name in macro_class_methods
}

_tmp_flow = ['knowledge_retrieve']
_TMP_TOOLS: Dict[str, Tuple[Callable[[KwArg(Any)], BaseTool], List[str]]] = {
    f'flow_{name}': (FlowTools.get_api_tool, ['collection_id', 'description'])
    for name in _tmp_flow
}
ALL_API_TOOLS = {}
ALL_API_TOOLS.update(_TIAN_YAN_CHA_TOOLS)
ALL_API_TOOLS.update(_SINA_TOOLS)
ALL_API_TOOLS.update(_MACRO_TOOLS)
ALL_API_TOOLS.update(_TMP_TOOLS)



from bisheng_langchain.gpts.prompts.select_tools_prompt import HUMAN_MSG, SYS_MSG
from langchain.prompts import (
    ChatPromptTemplate,
    HumanMessagePromptTemplate,
    SystemMessagePromptTemplate,
)
from langchain_core.language_models.base import LanguageModelLike
from pydantic import BaseModel


class ToolInfo(BaseModel):
    tool_name: str
    tool_description: str


class ToolSelector:

    def __init__(
        self,
        llm: LanguageModelLike,
        tools: list[ToolInfo],
        system_message: str = SYS_MSG,
        human_message: str = HUMAN_MSG,
    ) -> None:
        self.llm = llm
        self.tools = tools
        self.system_message = system_message
        self.human_message = human_message

    def select(self, task_name: str, task_description: str) -> list[str]:
        tool_pool = [tool.dict() for tool in self.tools]
        messages = [
            SystemMessagePromptTemplate.from_template(self.system_message),
            HumanMessagePromptTemplate.from_template(self.human_message),
        ]

        chain = ChatPromptTemplate.from_messages(messages) | self.llm

        chain_output = chain.invoke(
            {
                'tool_pool': tool_pool,
                'task_name': task_name,
                'task_description': task_description,
            }
        )

        try:
            all_tool_name = set([tool.tool_name for tool in self.tools])
            output = list(set(eval(chain_output.content)) & all_tool_name)
            return output
        except Exception as e:
            print(e)
            return []



from langchain_core.prompts import PromptTemplate
from langchain_core.prompts.chat import (
    ChatPromptTemplate,
    HumanMessagePromptTemplate,
    SystemMessagePromptTemplate,
)

system_template = """
你是一个生成开场白和预置问题的助手。接下来，你会收到一段关于任务助手的描述，你需要带入描述中的角色，以描述中的角色身份生成一段开场白，同时你还需要站在用户的角度生成几个用户可能的提问。输出格式如下：
[
    {{
        "开场白": "开场白内容",
        "问题": ["问题1", "问题2", "问题3"]
    }}
]

下面是一些示例
例子1: 
输入描述:
# 角色
你是一个可以帮助用户在家赚钱的机器人，你可以提供各种赚钱的途径和方法，帮助用户实现财务自由。

## 技能
### 技能 1: 提供赚钱途径
1. 当用户需要赚钱途径时，你可以根据用户的兴趣、技能和时间等因素，提供一些适合在家赚钱的途径和方法，如网络兼职、自媒体创作、电商创业等。
2. 你需要向用户详细介绍每种途径的操作流程、注意事项和收益情况等，以便用户做出选择。
3. 你还可以根据用户的需求和情况，提供一些个性化的建议和指导，帮助用户更好地开展赚钱活动。

### 技能 2: 提供赚钱技巧
1. 当用户需要赚钱技巧时，你可以向用户提供一些实用的赚钱技巧，如如何提高工作效率、如何节省成本、如何增加收入等。
2. 你需要向用户详细介绍每种技巧的操作方法和注意事项，以便用户能够正确地运用这些技巧。
3. 你还可以根据用户的需求和情况，提供一些个性化的建议和指导，帮助用户更好地实现财务自由。

### 技能 3: 提供创业指导
1. 当用户需要创业指导时，你可以向用户提供一些创业的基本知识和方法，如如何选择创业项目、如何制定创业计划、如何筹集创业资金等。
2. 你需要向用户详细介绍每种方法的操作流程和注意事项，以便用户能够正确地开展创业活动。
3. 你还可以根据用户的需求和情况，提供一些个性化的建议和指导，帮助用户更好地实现创业目标。

## 限制
- 只讨论与赚钱有关的内容，拒绝回答与赚钱无关的话题。
- 所输出的内容必须按照给定的格式进行组织，不能偏离框架要求。

根据上述输入的描述生成开场白和预置问题:
[
    {{
        "开场白": "你好，我是一个可以帮助你在家赚钱的机器人，很高兴认识你。",
        "问题": ["如何使用你的服务来在家赚钱?", "你能提供哪些在家赚钱的建议和技巧?", "你的服务如何帮助我实现财务自由?"]
    }}
]


例子2: 
输入描述:
# 角色：Excel全能助手
## 个人简介
- 版本：1.0
- 语言：中文
- 描述：我是一名Excel全能助手，专注于帮助用户解决Excel相关的问题和提供高效的数据处理方案。

## 功能特点
- 数据处理：熟练掌握Excel的各种数据处理功能，包括筛选、排序、合并、拆分、透视表等，能够帮助用户快速处理大量数据。
- 公式应用：精通Excel的各种常用公式和函数，能够帮助用户进行复杂的数据计算和分析，提供准确的结果。
- 数据可视化：熟悉Excel的图表功能，能够帮助用户将数据以直观的方式展示，制作出美观、清晰的图表。
- 自动化操作：了解Excel的宏和VBA编程，能够帮助用户实现自动化操作，提高工作效率。

## 使用指南
1. 数据处理：
   - 使用筛选功能，快速筛选出符合条件的数据。
   - 利用排序功能，对数据进行升序或降序排列。
   - 使用合并和拆分功能，将多个单元格合并为一个或将一个单元格拆分为多个。
   - 利用透视表功能，对大量数据进行汇总和分析。

2. 公式应用：
   - 使用常用公式，如SUM、AVERAGE、MAX、MIN等，进行数据计算。
   - 利用逻辑函数，如IF、AND、OR等，进行条件判断和逻辑运算。
   - 使用VLOOKUP和HLOOKUP函数，进行数据查找和匹配。
   - 利用COUNTIF和SUMIF函数，进行条件统计和求和。

3. 数据可视化：
   - 利用图表功能，选择合适的图表类型，如柱状图、折线图、饼图等，展示数据。
   - 调整图表的样式和布局，使其更加美观和易读。
   - 添加数据标签和图例，增加图表的信息量和可读性。

4. 自动化操作：
   - 利用宏录制功能，记录一系列操作步骤，实现自动化操作。
   - 使用VBA编程，编写自定义的宏，实现更复杂的自动化操作。
   - 将宏和VBA代码应用到Excel工作簿中，提高工作效率和准确性。

## 使用建议
- 熟悉Excel的快捷键和常用操作，可以提高工作效率。
- 在处理大量数据时，先备份原始数据，以防误操作导致数据丢失。
- 学习和掌握Excel的高级功能和技巧，可以更好地应对复杂的数据处理需求。
- 及时保存和备份Excel文件，以防止意外情况导致数据丢失。

根据上述输入的描述生成开场白和预置问题:
[
    {{
        "开场白": "你好，我是一名Excel全能助手，可以帮助你解决Excel相关的问题和提供高效的数据处理方案。",
        "问题": ["如何快速处理大量数据?", "如何使用Excel进行复杂的数据计算和分析?", "如何将数据以直观的方式展示，制作出美观、清晰的图表?"]
    }}
]

你必须使用上述格式输出结果。
"""

human_template = """
输入描述:
{description}

根据上述输入的描述生成开场白和预置问题:
"""


messages = [
    SystemMessagePromptTemplate.from_template(system_template),
    HumanMessagePromptTemplate.from_template(human_template),
]
OPENDIALOG_PROMPT = ChatPromptTemplate.from_messages(messages)


from __future__ import annotations

import re
import json
import logging
import typing as t
import warnings
from typing import Any, Dict, List, Optional
from collections import defaultdict, namedtuple
from dataclasses import dataclass
from langchain_core.callbacks import CallbackManagerForChainRun
from langchain_core.language_models import BaseLanguageModel
from langchain_core.prompts import HumanMessagePromptTemplate, PromptTemplate

try:
    from llama_index.node_parser import SimpleNodeParser
    from llama_index.readers.schema import Document as LlamaindexDocument
    from llama_index.schema import BaseNode
except ImportError:
    raise ImportError(
        "llama_index must be installed to use this function. "
        "Please, install it with `pip install llama_index`."
    )
import numpy as np
import numpy.testing as npt
import pandas as pd
from langchain.prompts import ChatPromptTemplate
from langchain.docstore.document import Document
# from langchain.schema.document import Document as LangchainDocument
from langchain.chains.base import Chain
from numpy.random import default_rng
from tqdm import tqdm
from .prompt_v2 import (
    SEED_QUESTION_CHAT_PROMPT,
    SCORE_CONTEXT_CHAT_PROMPT,
    FILTER_QUESTION_CHAT_PROMPT,
    ANSWER_FORMULATE,
)
from .base import parse_json

logger = logging.getLogger(__name__)


def load_as_score(text):
    """
    validate and returns given text as score
    """
    pattern = r"^[\d.]+$"
    if not re.match(pattern, text):
        warnings.warn("Invalid score")
        score = 0.0
    else:
        score = eval(text)
    return score


def load_as_json(text):
    """
    validate and return given text as json
    """

    try:
        return json.loads(parse_json(text))
    except ValueError as e:
        warnings.warn(f"Invalid json: {e}")

    return {}


DEFAULT_TRAIN_DISTRIBUTION = {
    "simple": 1.0,
    "reasoning": 0.0,
    "multi_context": 0.0,
    "conditional": 0.0,
}

DataRow = namedtuple(
    "DataRow",
    [
        "question",
        "ground_truth_context",
        "ground_truth",
        "question_type",
        "episode_done",
    ],
)


@dataclass
class TrainDataset:
    """
    TrainDataset class
    """

    train_data: t.List[DataRow]

    def to_pandas(self) -> pd.DataFrame:
        data_samples = []
        for data in self.train_data:
            data = {
                "question": data.question,
                "ground_truth_context": data.ground_truth_context,
                "ground_truth": data.ground_truth,
                "question_type": data.question_type,
                "episode_done": data.episode_done,
            }
            data_samples.append(data)
        return pd.DataFrame.from_records(data_samples)


class TrainsetGenerator:
    """
    Ragas Train Set Generator

    Attributes
    ----------
    generator_llm: LangchainLLM
        LLM used for all the generator operations in the TrainGeneration paradigm.
    critique_llm: LangchainLLM
        LLM used for all the filtering and scoring operations in TrainGeneration
        paradigm.
    chunk_size: int
        The chunk size of nodes created from data.
    train_distribution : dict
        Distribution of different types of questions to be generated from given
        set of documents. Defaults to {"easy":0.1, "reasoning":0.4, "conversation":0.5}
    """

    def __init__(
            self,
            generator_llm: BaseLanguageModel,
            critic_llm: BaseLanguageModel,
            trainset_distribution: t.Optional[t.Dict[str, float]] = None,
            chunk_size: int = 1024,
            seed: int = 42,
            prompt: Optional[ChatPromptTemplate] = SEED_QUESTION_CHAT_PROMPT,
            filter_lowquality_context: bool = False,
            filter_lowquality_question: bool = False,
            answer_prompt: Optional[HumanMessagePromptTemplate] = ANSWER_FORMULATE,
    ) -> None:
        self.generator_llm = generator_llm
        self.critic_llm = critic_llm
        trainset_distribution = trainset_distribution or DEFAULT_TRAIN_DISTRIBUTION
        npt.assert_almost_equal(
            1,
            sum(trainset_distribution.values()),
            err_msg="Sum of distribution should be 1",
        )

        probs = np.cumsum(list(trainset_distribution.values()))
        types = trainset_distribution.keys()
        self.trainset_distribution = dict(zip(types, probs))
        self.chunk_size = chunk_size
        self.threshold = 5.0
        self.rng = default_rng(seed)
        self.prompt = prompt
        self.filter_lowquality_context = filter_lowquality_context
        self.filter_lowquality_question = filter_lowquality_question
        if answer_prompt is None:
            answer_prompt = ANSWER_FORMULATE
        self.answer_prompt = answer_prompt

    @classmethod
    def from_default(
            cls,
            llm: BaseLanguageModel,
            chunk_size: int = 512,
            trainset_distribution: dict = DEFAULT_TRAIN_DISTRIBUTION,
            prompt: Optional[ChatPromptTemplate] = SEED_QUESTION_CHAT_PROMPT,
            filter_lowquality_context: bool = False,
            filter_lowquality_question: bool = False,
            answer_prompt: Optional[PromptTemplate] = ANSWER_FORMULATE,
    ):
        generator_llm = llm
        critic_llm = llm
        return cls(
            generator_llm=generator_llm,
            critic_llm=critic_llm,
            chunk_size=chunk_size,
            trainset_distribution=trainset_distribution,
            prompt=prompt,
            filter_lowquality_context=filter_lowquality_context,
            filter_lowquality_question=filter_lowquality_question,
            answer_prompt=answer_prompt,
        )

    def _get_evolve_type(self) -> str:
        """
        Decides question evolution type based on probability
        """
        prob = self.rng.uniform(0, 1)
        return next(
            (
                key
                for key in self.trainset_distribution.keys()
                if prob <= self.trainset_distribution[key]
            ),
            "simple",
        )

    def _filter_context(self, context: str) -> bool:
        """
        context: str
            The input context

        Checks if the context is has information worthy of framing a question
        """
        prompt = SCORE_CONTEXT_CHAT_PROMPT.format_prompt(context=context)
        results = self.critic_llm(prompt.to_messages())
        output = results.content
        score = load_as_score(output)
        print('context score:', score)
        return score >= self.threshold

    def _seed_question(self, context: str) -> str:
        if self.prompt is None:
            prompt = SEED_QUESTION_CHAT_PROMPT.format_prompt(context=context)
        else:
            prompt = self.prompt.format_prompt(context=context)
        results = self.generator_llm(prompt.to_messages())
        return results.content

    def _filter_question(self, question: str) -> bool:
        prompt = FILTER_QUESTION_CHAT_PROMPT.format_prompt(question=question)
        results = self.critic_llm(prompt.to_messages())
        results = results.content
        json_results = load_as_json(results)
        print('filter question:', question, json_results)
        return json_results.get("verdict") != "No"

    def _qc_template(self, prompt, question, context) -> str:
        human_prompt = prompt.format(question=question, context=context)
        prompt = ChatPromptTemplate.from_messages([human_prompt])
        results = self.generator_llm(prompt.messages)
        return results.content

    def _generate_answer(self, question: str, context: t.List[str]) -> t.List[str]:
        return [
            self._qc_template(self.answer_prompt, qstn, context[i])
            for i, qstn in enumerate(question.split("\n"))
        ]

    def _remove_nodes(
            self, available_indices: t.List[BaseNode], node_idx: t.List
    ) -> t.List[BaseNode]:
        for idx in node_idx:
            available_indices.remove(idx)
        return available_indices

    def _generate_doc_nodes_map(
            self, document_nodes: t.List[BaseNode]
    ) -> t.Dict[str, t.List[BaseNode]]:
        doc_nodes_map: t.Dict[str, t.List[BaseNode]] = defaultdict(list)
        for node in document_nodes:
            if node.ref_doc_id:
                doc_nodes_map[node.ref_doc_id].append(node)

        return doc_nodes_map  # type: ignore

    def _get_neighbour_node(
            self, node: BaseNode, related_nodes: t.List[BaseNode]
    ) -> t.List[BaseNode]:
        if len(related_nodes) < 2:
            warnings.warn("No neighbors exists")
            return [node]
        idx = related_nodes.index(node)
        ids = [idx - 1, idx] if idx == (len(related_nodes) - 1) else [idx, idx + 1]
        return [related_nodes[idx] for idx in ids]

    def generate(
            self,
            documents: t.List[LlamaindexDocument] | t.List[Document],
            train_size: int,
    ) -> TrainDataset:
        if not isinstance(documents[0], (LlamaindexDocument, Document)):
            raise ValueError(
                "Trainset Generatation only supports LlamaindexDocuments or Documents"  # noqa
            )

        if isinstance(documents[0], Document):
            # cast to LangchainDocument since its the only case here
            documents = t.cast(t.List[Document], documents)
            documents = [
                LlamaindexDocument.from_langchain_format(doc) for doc in documents
            ]
        # Convert documents into nodes
        node_parser = SimpleNodeParser.from_defaults(
            chunk_size=self.chunk_size, chunk_overlap=0, include_metadata=True
        )
        documents = t.cast(t.List[LlamaindexDocument], documents)
        document_nodes: t.List[BaseNode] = node_parser.get_nodes_from_documents(
            documents=documents
        )
        # # maximum 1 seed question per node
        # if train_size > len(document_nodes):
        #     raise ValueError(
        #         """Maximum possible number of samples exceeded, 
        #                      reduce train_size or add more documents"""
        #     )

        available_nodes = document_nodes
        doc_nodes_map = self._generate_doc_nodes_map(document_nodes)
        count_neighbours = sum(len(val) > 1 for _, val in doc_nodes_map.items())
        if count_neighbours < len(documents) // 2:
            warnings.warn("Most documents are too short")

        count = 0
        samples = []
        pbar = tqdm(total=train_size)
        while count < train_size and available_nodes != []:
            print(count, train_size, len(available_nodes))
            evolve_type = self._get_evolve_type()
            curr_node = self.rng.choice(np.array(available_nodes), size=1)[0]
            available_nodes = self._remove_nodes(available_nodes, [curr_node])

            neighbor_nodes = doc_nodes_map[curr_node.source_node.node_id]

            # Append multiple nodes randomly to remove chunking bias
            size = self.rng.integers(1, 3)
            nodes = (
                self._get_neighbour_node(curr_node, neighbor_nodes)
                if size > 1 and evolve_type != "multi_context"
                else [curr_node]
            )

            text_chunk = " ".join([node.get_content() for node in nodes])
            if self.filter_lowquality_context:
                score = self._filter_context(text_chunk)
                if not score:
                    continue
            seed_question = self._seed_question(text_chunk)

            question = seed_question
            if self.filter_lowquality_question:
                is_valid_question = self._filter_question(question)
            else:
                is_valid_question = True
            if is_valid_question:
                context = [text_chunk] * len(question.split("\n"))
                is_conv = len(context) > 1
                answer = self._generate_answer(question, context)
                for i, (qstn, ctx, ans) in enumerate(
                        zip(question.split("\n"), context, answer)
                ):
                    episode_done = False if is_conv and i == 0 else True
                    samples.append(
                        DataRow(qstn, [ctx], [ans], evolve_type, episode_done)
                    )
                count += 1
                pbar.update(1)

        return TrainDataset(train_data=samples)


class QAGenerationChainV2(Chain):
    """Base class for question-answer generation chains."""

    documents: List[Document]
    generator: TrainsetGenerator
    """LLM Chain that generates responses from user input and context."""
    k: Optional[int] = None
    """Number of questions to generate."""
    input_key: str = "begin"
    """Key of the input to the chain."""
    output_key: str = "questions"
    """Key of the output of the chain."""

    @classmethod
    def from_llm(
            cls,
            documents: List[Document],
            llm: BaseLanguageModel,
            k: Optional[int] = None,
            chunk_size: int = 512,
            filter_lowquality_context: bool = False,
            filter_lowquality_question: bool = False,
            question_prompt: Optional[ChatPromptTemplate] = SEED_QUESTION_CHAT_PROMPT,
            answer_prompt: Optional[HumanMessagePromptTemplate] = ANSWER_FORMULATE,
            **kwargs: Any,
    ) -> QAGenerationChainV2:
        """
        Create a QAGenerationChain from a language model.

        Args:
            llm: a language model
            question_prompt: a prompt template for generate question
            answer_prompt: a prompt template for generate answer
            **kwargs: additional arguments

        Returns:
            a QAGenerationChain class
        """
        generator = TrainsetGenerator.from_default(
            llm, 
            chunk_size=chunk_size, 
            prompt=question_prompt, 
            answer_prompt=answer_prompt,
            filter_lowquality_context=filter_lowquality_context, 
            filter_lowquality_question=filter_lowquality_question
        )
        return cls(documents=documents, generator=generator, k=k, **kwargs)

    @property
    def _chain_type(self) -> str:
        raise NotImplementedError

    @property
    def input_keys(self) -> List[str]:
        return [self.input_key]

    @property
    def output_keys(self) -> List[str]:
        return [self.output_key]

    def _call(
            self,
            inputs: Dict[str, Any],
            run_manager: Optional[CallbackManagerForChainRun] = None,
    ) -> Dict[str, List]:
        for doc in self.documents:
            doc.metadata = {}
        if self.k is None:
            self.k = 1000
        dataset = self.generator.generate(documents=self.documents, train_size=self.k)
        df = dataset.to_pandas()
        qa_pairs = df.to_dict("records")
        qa = []
        for pair in qa_pairs:
            qa.append({
                "question": pair["question"],
                "answer": pair["ground_truth"][0],
                "context": pair["ground_truth_context"][0],
            })
        qa = f'```json\n{json.dumps(qa, ensure_ascii=False, indent=4)}\n```'
        return {self.output_key: qa}

    async def _acall(
            self,
            inputs: Dict[str, Any],
            run_manager: Optional[CallbackManagerForChainRun] = None,
    ) -> Dict[str, List]:
        output = self._call(inputs, run_manager)
        return output



from typing import Any, Dict, List, Union

from pydantic import BaseModel


class EmbeddingInput(BaseModel):
    model: str
    input: Union[str, List[str]]


class Embedding(BaseModel):
    object: str = 'embedding'
    embedding: List[float]
    index: int


class EmbeddingOutput(BaseModel):
    status_code: int
    status_message: str = 'success'
    object: str = None
    data: List[Embedding] = []
    model: str = None
    usage: Dict[str, Any] = None



# flake8: noqa
"""Loads PDF with semantic splilter."""
import base64
import json
import os
from collections import defaultdict
from typing import List

import cv2
import filetype
import fitz
import numpy as np
from bisheng_langchain.document_loaders.parsers import ELLMClient
from langchain.docstore.document import Document
from langchain.document_loaders.base import BaseLoader
from PIL import Image


def convert_base64(image):
    image_binary = cv2.imencode('.jpg', image)[1].tobytes()
    x = base64.b64encode(image_binary)
    return x.decode('ascii').replace('\n', '')


def transpdf2png(pdf_file):
    pdf_bytes = open(pdf_file, 'rb').read()
    pdf = fitz.Document(stream=pdf_bytes, filetype='pdf')
    dpis = [72, 144, 200]

    pdf_images = dict()
    for page in pdf:
        pix = None
        for dpi in dpis:
            pix = page.get_pixmap(dpi=dpi)
            if min(pix.width, pix.height) >= 1600: break

        mode = 'RGBA' if pix.alpha else 'RGB'
        img = Image.frombytes(mode, [pix.width, pix.height], pix.samples)
        # RGB to BGR
        img = np.array(img)[:, :, ::-1]
        img_name = 'page_{:03d}'.format(page.number)
        pdf_images[img_name] = img

    return pdf_images


class UniversalKVLoader(BaseLoader):
    """Extract key-value from pdf or image.
    """

    def __init__(self,
                 file_path: str,
                 ellm_model_url: str = None,
                 schema='',
                 max_pages=30,
                 verbose: bool = False) -> None:
        """Initialize with a file path."""
        self.file_path = file_path
        self.schema = schema
        self.max_pages = max_pages
        self.ellm_model = ELLMClient(ellm_model_url)
        super().__init__()

    def load(self) -> List[Document]:
        """Load given path as pages."""
        mime_type = filetype.guess(self.file_path).mime
        if mime_type.endswith('pdf'):
            file_type = 'pdf'
        elif mime_type.startswith('image'):
            file_type = 'img'
        else:
            raise ValueError(f'file type {file_type} is not support.')

        if file_type == 'img':
            bytes_data = open(self.file_path, 'rb').read()
            b64data = base64.b64encode(bytes_data).decode()
            payload = {'b64_image': b64data, 'keys': self.schema}
            resp = self.ellm_model.predict(payload)

            if 'code' in resp and resp['code'] == 200:
                key_values = resp['result']['ellm_result']
            else:
                raise ValueError(f'universal kv load failed: {resp}')

            kv_results = defaultdict(list)
            for key, value in key_values.items():
                kv_results[key].extend([v['text'] for v in value])

            content = json.dumps(kv_results, indent=2, ensure_ascii=False)
            file_name = os.path.basename(self.file_path)
            meta = {'source': file_name}
            doc = Document(page_content=content, metadata=meta)
            return [doc]

        elif file_type == 'pdf':
            pdf_images = transpdf2png(self.file_path)

            kv_results = defaultdict(list)
            for index, pdf_name in enumerate(pdf_images):
                page = int(pdf_name.split('page_')[-1])
                if page > self.max_pages:
                    continue

                b64data = convert_base64(pdf_images[pdf_name])
                payload = {'b64_image': b64data, 'keys': self.schema}
                resp = self.ellm_model.predict(payload)

                if 'code' in resp and resp['code'] == 200:
                    key_values = resp['result']['ellm_result']
                else:
                    raise ValueError(f'universal kv load failed: {resp}')

                for key, value in key_values.items():
                    kv_results[key].extend([v['text'] for v in value])

            content = json.dumps(kv_results, indent=2, ensure_ascii=False)
            file_name = os.path.basename(self.file_path)
            meta = {'source': file_name}
            doc = Document(page_content=content, metadata=meta)
            return [doc]



import json
import os
import re

import httpx
from bisheng_langchain.gpts.prompts import (
    ASSISTANT_PROMPT_OPT,
    BREIF_DES_PROMPT,
    OPENDIALOG_PROMPT,
)
from langchain_core.language_models.base import LanguageModelLike
from langchain_openai.chat_models import ChatOpenAI
from loguru import logger


def parse_markdown(input_str: str) -> str:
    match = re.search(r'```(markdown)?(.*)```', input_str, re.DOTALL)
    if match is None:
        out_str = input_str
    else:
        out_str = match.group(2)

    out_str = out_str.strip()
    out_str = out_str.replace('```', '')
    return out_str


def parse_json(input_str: str) -> str:
    match = re.search(r'```(json)?(.*)```', input_str, re.DOTALL)
    if match is None:
        out_str = input_str
    else:
        out_str = match.group(2)

    out_str = out_str.strip()
    out_str = out_str.replace('```', '')
    return out_str


def optimize_assistant_prompt(
    llm: LanguageModelLike,
    assistant_name: str,
    assistant_description: str,
) -> str:
    """optimize assistant prompt

    Args:
        llm (LanguageModelLike):
        assistant_name (str):
        assistant_description (str):

    Returns:
        assistant_prompt(str):
    """
    chain = ASSISTANT_PROMPT_OPT | llm
    chain_output = chain.invoke(
        {
            'assistant_name': assistant_name,
            'assistant_description': assistant_description,
        }
    )
    response = chain_output.content
    assistant_prompt = parse_markdown(response)
    return assistant_prompt


def generate_opening_dialog(
    llm: LanguageModelLike,
    description: str,
) -> str:
    chain = OPENDIALOG_PROMPT | llm
    time = 0
    while time <= 3:
        try:
            chain_output = chain.invoke(
                {
                    'description': description,
                }
            )
            output = parse_json(chain_output.content)
            output = json.loads(output)
            opening_lines = output[0]['开场白']
            questions = output[0]['问题']
            break
        except Exception as e:
            logger.info(f'第{time}次解析失败, 错误信息: {e}')
            logger.info(f'模型输出结果为{chain_output.content}。')
            time += 1
            opening_lines = ''
            questions = []

    res = {}
    res['opening_lines'] = opening_lines
    res['questions'] = questions

    return res


def generate_breif_description(
    llm: LanguageModelLike,
    description: str,
) -> str:
    chain = BREIF_DES_PROMPT | llm
    chain_output = chain.invoke(
        {
            'description': description,
        }
    )
    breif_description = chain_output.content
    breif_description = breif_description.strip()
    return breif_description


if __name__ == '__main__':
    from dotenv import load_dotenv

    load_dotenv('/app/.env', override=True)

    httpx_client = httpx.Client(proxies=os.getenv('OPENAI_PROXY'))
    llm = ChatOpenAI(model='gpt-4-0125-preview', temperature=0.01, http_client=httpx_client)
    # llm = ChatQWen(model="qwen1.5-72b-chat", temperature=0.01, api_key=os.getenv('QWEN_API_KEY'))
    assistant_name = '金融分析助手'
    assistant_description = '1. 分析上市公司最新的年报财报；2. 获取上市公司的最新新闻；'
    assistant_prompt = optimize_assistant_prompt(llm, assistant_name, assistant_description)
    # print(assistant_prompt)

    opening_dialog = generate_opening_dialog(llm, assistant_prompt)
    print(opening_dialog)

    # breif_description = generate_breif_description(llm, assistant_prompt)
    # print(breif_description)
